# -*- coding: utf-8 -*-
"""
Functions and classes for generating analytics on physiological data
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import welch


class HRV:
    """
    Class for calculating various HRV statistics

    Parameters
    ----------
    data : Physio_like
        Physiological data object with detected peaks and troughs

    Attributes
    ----------
    rrint : :obj:`numpy.ndarray`
        R-R intervals derived from `data` (sometimes referred to as N-N
        intervals in derived metrics)
    rrtime : :obj:`numpy.ndarray`
        Time stamps of `rrint`
    avgnn : float
        Average heart rate (N-N interval)
    sdnn : float
        Standard deviation of heart rate (N-N intervals)
    rmssd : float
        Root mean square of successive differences
    sdsd : float
        Standard deviation of successive differences
    nn50 : float
        Number of N-N intervals greater than 50ms
    pnn50 : float
        Percent of N-N intervals greater than 50ms
    nn20 : float
        Number of N-N intervals greater than 20ms
    pnn20 : float
        Percent of N-N intervals greater than 20ms
    hf : float
        High-frequency power of R-R intervals, summed across 0.15-0.40 Hz
    hf_log : float
        Log of `hf`
    lf : float
        Low-frequency power of R-R intervals, summed across 0.04-0.15 Hz
    lf_log : float
        Log of `lf`
    vlf : float
        Very low frequency power of R-R intervals, summed across 0-0.04 Hz
    vlf_log : float
        Log of `vlf`
    lftohf : float
        Ratio of `lf` over `hf`
    hf_peak : float
        Peak frequency in `hf` band (0.15-0.40 Hz)
    lf_peak : float
        Peak frequency in `lf` band (0.04-0.15 Hz)

    Notes
    -----
    Uses scipy.signal.welch for calculation of frequency-based statistics
    """

    def __init__(self, data):
        self.data = data
        func = interp1d(self.rrtime, self.rrint * 1000, kind="cubic")
        irrt = np.arange(self.rrtime[0], self.rrtime[-1], 1.0 / 4.0)
        self._irri = func(irrt)

    @property
    def rrtime(self):
        """Times of R-R intervals (in seconds)"""
        if len(self.data.peaks):
            diff = (self.data._masked[:-1] + self.data._masked[1:]) / (2 * self.data.fs)
            return diff.compressed()

    @property
    def rrint(self):
        """Length of R-R intervals (in seconds)"""
        if len(self.data.peaks):
            return (np.diff(self.data._masked) / self.data.fs).compressed()

    @property
    def _sd(self):
        return np.diff(np.diff(self.data._masked)).compressed()

    @property
    def _fft(self):
        return welch(self._irri, nperseg=120, fs=4.0, scaling="spectrum")

    @property
    def avgnn(self):
        return self.rrint.mean() * 1000

    @property
    def sdnn(self):
        return self.rrint.std() * 1000

    @property
    def rmssd(self):
        return np.sqrt((self._sd**2).mean())

    @property
    def sdsd(self):
        return self._sd.std()

    @property
    def nn50(self):
        return np.argwhere(self._sd > 50.0).size

    @property
    def pnn50(self):
        return self.nn50 / self.rrint.size

    @property
    def nn20(self):
        return np.argwhere(self._sd > 20.0).size

    @property
    def pnn20(self):
        return self.nn20 / self.rrint.size

    @property
    def _hf(self):
        fx, px = self._fft
        return px[np.logical_and(fx >= 0.15, fx < 0.40)]

    @property
    def _lf(self):
        fx, px = self._fft
        return px[np.logical_and(fx >= 0.04, fx < 0.15)]

    @property
    def _vlf(self):
        fx, px = self._fft
        return px[np.logical_and(fx >= 0.0, fx < 0.04)]

    @property
    def hf(self):
        return sum(self._hf)

    @property
    def hf_log(self):
        return np.log(self.hf)

    @property
    def lf(self):
        return sum(self._lf)

    @property
    def lf_log(self):
        return np.log(self.lf)

    @property
    def vlf(self):
        return sum(self._vlf)

    @property
    def vlf_log(self):
        return np.log(self.vlf)

    @property
    def lftohf(self):
        return self.lf / self.hf

    @property
    def hf_peak(self):
        fx, px = self._fft
        return fx[np.argmax(self._hf)]

    @property
    def lf_peak(self):
        fx, px = self._fft
        return fx[np.argmax(self._lf)]



# -*- coding: utf-8 -*-

import numpy as np
import pytest

from peakdet.physio import Physio
from peakdet.tests import utils as testutils

DATA = np.loadtxt(testutils.get_test_data_path("ECG.csv"))
PROPERTIES = ["data", "fs", "history", "peaks", "troughs", "_masked"]
PHYSIO_TESTS = [
    # accepts "correct" inputs for history
    dict(kwargs=dict(data=DATA, history=[("good", "history")])),
    # fails on bad inputs for history
    dict(kwargs=dict(data=DATA, history=["malformed", "history"]), raises=TypeError),
    dict(kwargs=dict(data=DATA, history="not real history"), raises=TypeError),
    # accepts "correct" for metadata
    dict(kwargs=dict(data=DATA, metadata=dict())),
    dict(kwargs=dict(data=DATA, metadata=dict(peaks=[], reject=[], troughs=[]))),
    # fails on bad inputs for metadata
    dict(kwargs=dict(data=DATA, metadata=[]), raises=TypeError),
    dict(kwargs=dict(data=DATA, metadata=dict(peaks={})), raises=TypeError),
    # fails on bad inputs for data
    dict(kwargs=dict(data=np.column_stack([DATA, DATA])), raises=ValueError),
    dict(kwargs=dict(data="hello"), raises=ValueError),
]


def test_physio():
    phys = Physio(DATA, fs=1000)
    assert len(np.hstack((phys[:10], phys[10:-10], phys[-10:])))
    assert str(phys) == "Physio(size=44611, fs=1000.0)"
    assert len(np.exp(phys)) == 44611


class TestPhysio:
    tests = PHYSIO_TESTS

    def test_physio_creation(self):
        for test in PHYSIO_TESTS:
            if test.get("raises") is not None:
                with pytest.raises(test["raises"]):
                    phys = Physio(**test["kwargs"])
            else:
                phys = Physio(**test["kwargs"])
                for prop in PROPERTIES:
                    assert hasattr(phys, prop)
                for prop in ["peaks", "reject", "troughs"]:
                    assert isinstance(phys._metadata.get(prop), np.ndarray)


# TODO: Update unit test
def test_neurokit2phys(path_neurokit):
    df = pd.read_csv(path_neurokit, sep="\t")  # noqa
    phys = Physio.neurokit2phys(
        path_neurokit, copy_data=True, copy_peaks=True, copy_troughs=True, fs=fs  # noqa
    )

    assert all(
        np.unique(
            phys.data
            == np.hstack(df.loc[:, df.columns.str.endswith("Clean")].to_numpy())
        )
    )
    assert all(
        np.unique(
            phys.peaks == np.where(df.loc[:, df.columns.str.endswith("Peaks")] != 0)[0]
        )
    )
    assert phys.fs == fs  # noqa


# TODO: Update unit test
def test_phys2neurokit(path_phys):
    phys = load_physio(path_phys, allow_pickle=True)  # noqa
    neuro = data.phys2neurokit(  # noqa
        copy_data=True,
        copy_peaks=True,
        copy_troughs=False,
        module=module,  # noqa
        neurokit_path=path_neurokit,  # noqa
    )

    assert all(
        np.unique(
            phys.data
            == np.hstack(neuro.loc[:, neuro.columns.str.endswith("Raw")].to_numpy())
        )
    )
    assert all(
        np.unique(
            phys.peaks
            == np.where(neuro.loc[:, neuro.columns.str.endswith("Peaks")] != 0)[0]
        )
    )



# -*- coding: utf-8 -*-
"""
Helper class for holding physiological data and associated metadata information
"""

import numpy as np
from loguru import logger


class Physio:
    """
    Class to hold physiological data and relevant information

    Parameters
    ----------
    data : array_like
        Input data array
    fs : float, optional
        Sampling rate of `data` (Hz). Default: None
    history : list of tuples, optional
        Functions performed on `data`. Default: None
    metadata : dict, optional
        Metadata associated with `data`. Default: None
    suppdata : array_like, optional
        Support data array. Default: None

    Attributes
    ----------
    data : :obj:`numpy.ndarray`
        Physiological waveform
    fs : float
        Sampling rate of `data` in Hz
    history : list of tuples
        History of functions that have been performed on `data`, with relevant
        parameters provided to functions.
    peaks : :obj:`numpy.ndarray`
        Indices of peaks in `data`
    troughs : :obj:`numpy.ndarray`
        Indices of troughs in `data`
    suppdata : :obj:`numpy.ndarray`
        Secondary physiological waveform
    """

    def __init__(self, data, fs=None, history=None, metadata=None, suppdata=None):
        logger.debug("Initializing new Physio object")
        self._data = np.asarray(data).squeeze()
        if self.data.ndim > 1:
            raise ValueError(
                "Provided data dimensionality {} > 1.".format(self.data.ndim)
            )
        if not np.issubdtype(self.data.dtype, np.number):
            raise ValueError(
                "Provided data of type {} is not numeric.".format(self.data.dtype)
            )
        self._fs = np.float64(fs)
        self._history = [] if history is None else history
        if not isinstance(self._history, list) or any(
            [not isinstance(f, tuple) for f in self._history]
        ):
            raise TypeError(
                "Provided history {} must be a list-of-tuples. "
                "Please check inputs.".format(history)
            )
        if metadata is not None:
            if not isinstance(metadata, dict):
                raise TypeError(
                    "Provided metadata {} must be dict-like.".format(metadata)
                )
            for k in ["peaks", "troughs", "reject"]:
                metadata.setdefault(k, np.empty(0, dtype=int))
                if not isinstance(metadata.get(k), np.ndarray):
                    try:
                        metadata[k] = np.asarray(metadata.get(k), dtype=int)
                    except TypeError:
                        raise TypeError(
                            "Provided metadata must be dict-like"
                            "with integer array entries."
                        )
            self._metadata = dict(**metadata)
        else:
            self._metadata = dict(
                peaks=np.empty(0, dtype=int),
                troughs=np.empty(0, dtype=int),
                reject=np.empty(0, dtype=int),
            )
        self._suppdata = None if suppdata is None else np.asarray(suppdata).squeeze()

    def __array__(self):
        return self.data

    def __getitem__(self, slicer):
        return self.data[slicer]

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return "{name}(size={size}, fs={fs})".format(
            name=self.__class__.__name__, size=self.data.size, fs=self.fs
        )

    __repr__ = __str__

    @property
    def data(self):
        """Physiological data"""
        return self._data

    @property
    def fs(self):
        """Sampling rate of data (Hz)"""
        return self._fs

    @property
    def history(self):
        """Functions that have been performed on / modified `data`"""
        return self._history

    @property
    def peaks(self):
        """Indices of detected peaks in `data`"""
        return self._masked.compressed()

    @property
    def troughs(self):
        """Indices of detected troughs in `data`"""
        return self._metadata["troughs"]

    @property
    def _masked(self):
        return np.ma.masked_array(
            self._metadata["peaks"],
            mask=np.isin(self._metadata["peaks"], self._metadata["reject"]),
        )

    @property
    def suppdata(self):
        """Physiological data"""
        return self._suppdata

    def phys2neurokit(
        self, copy_data, copy_peaks, copy_troughs, module, neurokit_path=None
    ):
        """Physio to neurokit dataframe

        Parameters
        ----------
        copy_data: bool
            whether to copy raw data from Physio object to dataframe
        copy_peaks: bool
            whether to copy peaks from Physio object to dataframe
        copy_troughs: bool
            whether to copy troughs from Physio object to dataframe
        module: string
            name of module (eg. 'EDA', 'RSP', 'PPG'...)
        neurokit_path: string
            path to neurokit dataframe
        """
        import pandas as pd

        if neurokit_path is not None:
            df = pd.read_csv(neurokit_path, sep="\t")
        else:
            df = pd.DataFrame(
                0,
                index=np.arange(len(self.data)),
                columns=["%s_Raw" % module, "%s_Peaks" % module, "%s_Troughs" % module],
            )

        if copy_data:
            df.loc[:, df.columns.str.endswith("Raw")] = self.data

        if copy_peaks:
            b_peaks = np.zeros(len(self.data))
            b_peaks[self.peaks] = 1
            df.loc[:, df.columns.str.endswith("Peaks")] = b_peaks

        if copy_troughs:
            b_troughs = np.zeros(len(self.data))
            b_troughs[self.troughs] = 1
            df.loc[:, df.columns.str.endswith("Troughs")] = b_troughs

        return df

    @classmethod
    def neurokit2phys(
        cls, neurokit_path, fs, copy_data, copy_peaks, copy_troughs, **kwargs
    ):
        """Neurokit dataframe to phys

        Parameters
        ----------
        neurokit_path: string
            path to neurokit dataframe
        fs: float
            sampling rate
        copy_data: bool
            whether to copy raw data from Physio object to dataframe
        copy_peaks: bool
            whether to copy peaks from Physio object to dataframe
        copy_troughs: bool
            whether to copy troughs from Physio object to dataframe
        suppdata: array_like, optional
            Support data array. Default: None
        """
        import pandas as pd

        df = pd.read_csv(neurokit_path, sep="\t")

        if copy_data:
            # if cleaned data exists, substitute 'data' with cleaned data, else use raw data
            if df.columns.str.endswith("Clean").any():
                data = np.hstack(df.loc[:, df.columns.str.endswith("Clean")].to_numpy())
            elif df.columns.str.endswith("Raw").any():
                data = np.hstack(df.loc[:, df.columns.str.endswith("Raw")].to_numpy())

        if copy_peaks:
            # if peaks exists
            if df.columns.str.endswith("Peaks").any():
                peaks = np.where(df.loc[:, df.columns.str.endswith("Peaks")] == 1)[0]

        if copy_troughs:
            # if troughs exists
            if df.columns.str.endswith("Troughs").any():
                troughs = np.where(df.loc[:, df.columns.str.endswith("Troughs")] == 1)[
                    0
                ]

        if "peaks" in locals() and "troughs" in locals():
            metadata = dict(peaks=peaks, troughs=troughs)
        elif "peaks" in locals() and "troughs" not in locals():
            metadata = dict(peaks=peaks)

        return cls(data, fs=fs, metadata=metadata, **kwargs)



# -*- coding: utf-8 -*-

import numpy as np


class HRModality:
    def iHR(self, step=1, start=0, end=None, TR=None):
        if end is None:
            end = self.rrtime[-1]

        mod = self.TR * (step // 2)
        time = np.arange(start - mod, end + mod + 1, self.TR, dtype="int")
        HR = np.zeros(len(time) - step)

        for tpoint in range(step, time.size):
            inds = np.logical_and(
                self.rrtime >= time[tpoint - step], self.rrtime < time[tpoint]
            )
            relevant = self.rrint[inds]

            if relevant.size == 0:
                continue
            HR[tpoint - step] = (60 / relevant).mean()

        return HR

    def meanHR(self):
        return np.mean(60 / self.rrint)


class ECG:
    flims = [5, 15.0]


class PPG:
    flims = 2.0


class RESP:
    flims = [0.05, 0.5]

    def RVT(self, start=0, end=None, TR=None):
        if end is None:
            end = self.rrtime[-1]

        pheight, theight = self.data[self.peakinds], self.data[self.troughinds]
        rvt = (pheight[:-1] - theight) / (np.diff(self.peakinds) / self.fs)
        rt = (self.peakinds / self.fs)[1:]

        time = np.arange(start, end + 1, self.TR, dtype="int")
        iRVT = np.interp(time, rt, rvt, left=rvt.mean(), right=rvt.mean())

        return iRVT



# -*- coding: utf-8 -*-

import json
import os

import numpy as np
import pytest

from peakdet import io, operations, physio
from peakdet.tests.utils import get_test_data_path


def test_load_physio(caplog):
    # try loading pickle file (from io.save_physio)
    pckl = io.load_physio(get_test_data_path("ECG.phys"), allow_pickle=True)
    assert isinstance(pckl, physio.Physio)
    assert pckl.data.size == 44611
    assert pckl.fs == 1000.0
    pckl = io.load_physio(get_test_data_path("ECG.phys"), fs=500.0, allow_pickle=True)
    assert caplog.text.count("WARNING") == 1
    assert pckl.fs == 500.0

    # try loading CSV file
    csv = io.load_physio(get_test_data_path("ECG.csv"))
    assert isinstance(csv, physio.Physio)
    assert np.allclose(csv, pckl)
    assert np.isnan(csv.fs)
    assert csv.history[0][0] == "load_physio"

    # try loading array
    arr = io.load_physio(np.loadtxt(get_test_data_path("ECG.csv")))
    assert caplog.text.count("WARNING") == 2
    assert isinstance(arr, physio.Physio)
    arr = io.load_physio(
        np.loadtxt(get_test_data_path("ECG.csv")),
        history=[("np.loadtxt", {"fname": "ECG.csv"})],
    )
    assert isinstance(arr, physio.Physio)

    # try loading physio object (and resetting dtype)
    out = io.load_physio(arr, dtype="float32")
    assert out.data.dtype == np.dtype("float32")
    assert out.history[0][0] == "np.loadtxt"
    assert out.history[-1][0] == "load_physio"
    with pytest.raises(TypeError):
        io.load_physio([1, 2, 3])


def test_save_physio(tmpdir):
    pckl = io.load_physio(get_test_data_path("ECG.phys"), allow_pickle=True)
    out = io.save_physio(tmpdir.join("tmp").purebasename, pckl)
    assert os.path.exists(out)
    assert isinstance(io.load_physio(out, allow_pickle=True), physio.Physio)


def test_load_history(tmpdir):
    # get paths of data, new history
    fname = get_test_data_path("ECG.csv")
    temp_history = tmpdir.join("tmp").purebasename

    # make physio object and perform some operations
    phys = io.load_physio(fname, fs=1000.0)
    filt = operations.filter_physio(phys, [5.0, 15.0], "bandpass")

    # save history to file and recreate new object from history
    path = io.save_history(temp_history, filt)
    replayed = io.load_history(path, verbose=True)

    # ensure objects are the same
    assert np.allclose(filt, replayed)
    assert filt.history == replayed.history
    assert filt.fs == replayed.fs


def test_save_history(tmpdir, caplog):
    # get paths of data, original history, new history
    fname = get_test_data_path("ECG.csv")
    orig_history = get_test_data_path("history.json")
    temp_history = tmpdir.join("tmp").purebasename

    # make physio object and perform some operations
    phys = physio.Physio(np.loadtxt(fname), fs=1000.0)
    io.save_history(temp_history, phys)
    assert caplog.text.count("WARNING") == 1  # no history = warning
    filt = operations.filter_physio(phys, [5.0, 15.0], "bandpass")
    path = io.save_history(temp_history, filt)  # dump history=

    # load both original and new json and ensure equality
    with open(path, "r") as src:
        hist = json.load(src)
    with open(orig_history, "r") as src:
        orig = json.load(src)
    assert hist == orig



# -*- coding: utf-8 -*-

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from peakdet import physio, utils
from peakdet.tests import utils as testutils

DATA, PEAKS, TROUGHS = testutils.get_sample_data()
GET_CALL_ARGUMENTS = [
    # check basic functionality
    dict(
        function="get_call_func",
        input=dict(arg1=1, arg2=2, serializable=False),
        expected=dict(arg1=1, arg2=2, kwarg1=10, kwarg2=20),
    ),
    # check that changing kwargs in function persists to output
    dict(
        function="get_call_func",
        input=dict(arg1=11, arg2=21, serializable=False),
        expected=dict(arg1=11, arg2=21, kwarg1=21, kwarg2=41),
    ),
    # confirm serializability is effective
    dict(
        function="get_call_func",
        input=dict(arg1=1, arg2=2, kwarg1=np.array([1, 2, 3]), serializable=True),
        expected=dict(arg1=1, arg2=2, kwarg1=[1, 2, 3], kwarg2=20),
    ),
]


def test_get_call():
    for entry in GET_CALL_ARGUMENTS:
        fcn, args = testutils.get_call_func(**entry["input"])
        assert fcn == entry["function"]
        assert args == entry["expected"]


def test_check_physio():
    fname = testutils.get_test_data_path("ECG.csv")
    data = physio.Physio(np.loadtxt(fname), fs=1000.0)
    # check that `ensure_fs` is functional
    with pytest.raises(ValueError):
        utils.check_physio(fname)
    assert isinstance(utils.check_physio(fname, False), physio.Physio)
    # "normal" instance should just pass
    assert isinstance(utils.check_physio(data), physio.Physio)
    # check that copy works
    assert utils.check_physio(data) == data
    assert utils.check_physio(data, copy=True) != data


def test_new_physio_like():
    fname = testutils.get_test_data_path("ECG.csv")
    data = physio.Physio(np.loadtxt(fname), fs=1000.0)
    data._history = [("does history", "copy?")]
    data._metadata["peaks"] = np.array([1, 2, 3])
    # assert all copies happen by default
    new_data = utils.new_physio_like(data, data[:])
    assert np.allclose(data, utils.new_physio_like(data, data[:]))
    assert new_data.fs == data.fs
    assert new_data.data.dtype == data.data.dtype
    assert new_data.history == data.history
    assert new_data._metadata == data._metadata
    # check if changes apply
    new_data = utils.new_physio_like(
        data, data[:], fs=50, dtype=int, copy_history=False, copy_metadata=False
    )
    assert np.allclose(data, utils.new_physio_like(data, data[:]))
    assert new_data.fs == 50
    assert new_data.data.dtype == int
    assert new_data.history == []
    for k, v in new_data._metadata.items():
        assert v.size == 0


def test_check_troughs():
    true = np.array([9, 21])
    assert_array_equal(utils.check_troughs(DATA, PEAKS), true)



# -*- coding: utf-8 -*-
import argparse
import datetime
import glob
import os
import sys

from loguru import logger

import peakdet

TARGET = "pythonw" if sys.platform == "darwin" else "python"
TARGET += " -u " + os.path.abspath(__file__)

LOADERS = dict(rtpeaks=peakdet.load_rtpeaks, MRI=peakdet.load_physio)

MODALITIES = dict(
    ECG=([5.0, 15.0], "bandpass"), PPG=(2, "lowpass"), RESP=([0.05, 0.5], "bandpass")
)

ATTR_CONV = {
    "Average NN intervals": "avgnn",
    "Root mean square of successive differences": "rmssd",
    "Standard deviation of NN intervals": "sdnn",
    "Standard deviation of successive differences": "sdsd",
    "Number of successive differences >50 ms": "nn50",
    "Percent of successive differences >50 ms": "pnn50",
    "Number of successive differences >20 ms": "nn20",
    "Percent of successive differences >20 ms": "pnn20",
    "High frequency HRV hfHRV": "hf",
    "Log of high frequency HRV, log(hfHRV)": "hf_log",
    "Low frequency HRV, lfHRV": "lf",
    "Log of low frequency HRV, log(lfHRV)": "lf_log",
    "Very low frequency HRV, vlfHRV": "vlf",
    "Log of very low frequency HRV, log(vlfHRV)": "vlf_log",
    "Ratio of lfHRV : hfHRV": "lftohf",
    "Peak frequency of hfHRV": "hf_peak",
    "Peak frequency of lfHRV": "lf_peak",
}


def get_parser():
    """Parser for GUI and command-line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file_template",
        help="Select a representative file and replace all "
        'subject-specific information with a "?" symbol.'
        "\nFor example, subject_001_data.txt should "
        "become subject_???_data.txt and will expand to "
        "match\nsubject_001_data.txt, subject_002_data."
        "txt, ..., subject_999_data.txt.",
    )

    inp_group = parser.add_argument_group(
        "Inputs", "Options to specify " "format of input files"
    )
    inp_group.add_argument(
        "--modality",
        default="ECG",
        choices=list(MODALITIES.keys()),
        help="Modality of input data.",
    )
    inp_group.add_argument(
        "--fs",
        default=1000.0,
        type=float,
        help="Sampling rate of input data.",
    )
    inp_group.add_argument(
        "--source",
        default="rtpeaks",
        choices=list(LOADERS.keys()),
        help="Program used to collect the data.",
    )
    inp_group.add_argument(
        "--channel",
        default=1,
        type=int,
        help="Which channel of data to read from data "
        'files.\nOnly applies if "Source" is set to '
        "rtpeaks.",
    )

    out_group = parser.add_argument_group(
        "Outputs", "Options to specify " "format of output files"
    )
    out_group.add_argument(
        "-o",
        "--output",
        default="peakdet.csv",
        help="Output filename for generated measurements.",
    )
    out_group.add_argument(
        "-m",
        "--measurements",
        metavar="Measurements",
        nargs="+",
        choices=list(ATTR_CONV.keys()),
        default=["Average NN intervals", "Standard deviation of NN intervals"],
        help="Desired physiological measurements",
    )
    out_group.add_argument(
        "-s",
        "--savehistory",
        action="store_true",
        help="Whether to save history of data processing " "for each file.",
    )

    edit_group = parser.add_argument_group(
        "Workflow arguments (optional!)",
        "Options to specify modifications " "to workflow",
    )
    edit_group.add_argument(
        "-n",
        "--noedit",
        action="store_true",
        help="Turn off interactive editing.",
    )
    edit_group.add_argument(
        "-t",
        "--thresh",
        default=0.2,
        type=float,
        help="Threshold for peak detection algorithm.",
    )

    log_style_group = parser.add_argument_group(
        "Logging style arguments (optional and mutually exclusive)",
        "Options to specify the logging style",
    )
    log_style_group_exclusive = log_style_group.add_mutually_exclusive_group()
    log_style_group_exclusive.add_argument(
        "-debug",
        "--debug",
        dest="debug",
        action="store_true",
        help="Print additional debugging info and error diagnostics to log file. Default is False.",
        default=False,
    )
    log_style_group_exclusive.add_argument(
        "-quiet",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="Only print warnings to log file. Default is False.",
        default=False,
    )

    return parser


@logger.catch
def workflow(
    *,
    file_template,
    modality,
    fs,
    source="MRI",
    channel=1,
    output="peakdet.csv",
    savehistory=True,
    noedit=False,
    thresh=0.2,
    measurements=ATTR_CONV.keys(),
    debug=False,
    quiet=False,
):
    """
    Basic workflow for physiological data

    Parameters
    ----------
    file_template : str
        Template filename for data inputs
    modality : {'ECG', 'PPG', 'RESP'}
        Currently support data modalities
    fs : float
        Sampling rate of input data
    source : {'rtpeaks', 'MRI'}, optional
        How data were acquired. Default: 'MRI'
    channel : int, optional
        Which channel of data to analyze; only applies if source is 'rtpeaks'.
        Default: 1
    output : str, optional
        Desired output filename. Default: 'peakdet.csv'
    savehistory : bool, optional
        Whether to save editing history of each file with
        ``peakdet.save_history``. History will be used if this workflow is
        run again on the samed data files. Default: True
    noedit : bool, optional
        Whether to disable interactive editing of physio data. Default: False
    thresh : [0, 1] float, optional
        Threshold for peak detection. Default: 0.2
    measurements : list, optional
        Which HRV-related measurements to save from data. See ``peakdet.HRV``
        for available measurements. Default: all available measurements.
    verbose : bool, optional
        Whether to include verbose logs when catching exceptions that include diagnostics
    """
    outdir = os.path.dirname(output)
    logger.info(f"Current path is {outdir}")

    # Create logfile name
    basename = "peakdet"
    extension = "log"
    isotime = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
    logname = os.path.join(outdir, (basename + isotime + "." + extension))

    logger.remove(0)
    if quiet:
        logger.add(
            sys.stderr,
            level="WARNING",
            colorize=True,
            backtrace=False,
            diagnose=False,
        )
        logger.add(
            logname,
            level="WARNING",
            colorize=False,
            backtrace=False,
            diagnose=False,
        )
    elif debug:
        logger.add(
            sys.stderr,
            level="DEBUG",
            colorize=True,
            backtrace=True,
            diagnose=True,
        )
        logger.add(
            logname,
            level="DEBUG",
            colorize=False,
            backtrace=True,
            diagnose=True,
        )
    else:
        logger.add(
            sys.stderr,
            level="INFO",
            colorize=True,
            backtrace=True,
            diagnose=False,
        )
        logger.add(
            logname,
            level="INFO",
            colorize=False,
            backtrace=True,
            diagnose=False,
        )

    # output file
    logger.info("OUTPUT FILE:\t\t{}".format(output))
    # grab files from file template
    logger.info("FILE TEMPLATE:\t{}".format(file_template))
    files = glob.glob(file_template, recursive=True)

    # convert measurements to peakdet.HRV attribute friendly names
    try:
        logger.info("REQUESTED MEASUREMENTS: {}\n".format(", ".join(measurements)))
    except TypeError:
        raise TypeError(
            "It looks like you didn't select any of the options "
            "specifying desired output measurements. Please "
            "select at least one measurement and try again."
        )
    measurements = [ATTR_CONV[attr] for attr in measurements]

    # get appropriate loader
    load_func = LOADERS[source]

    # check if output file exists -- if so, ensure headers will match
    head = "filename," + ",".join(measurements)
    if os.path.exists(output):
        with open(output, "r") as src:
            eheader = src.readlines()[0]
        # if existing output file does not have same measurements are those
        # requested on command line, warn and use existing measurements so
        # as not to totally fork up existing file
        if eheader != head:
            logger.warning(
                "Desired output file already exists and requested "
                "measurements do not match with measurements in "
                "existing output file. Using the pre-existing "
                "measurements, instead."
            )
            measurements = [f.strip() for f in eheader.split(",")[1:]]
        head = ""
    # if output file doesn't exist, nbd
    else:
        head += "\n"

    with open(output, "a+") as dest:
        dest.write(head)
        # iterate through all files and do peak detection with manual editing
        for fname in files:
            fname = os.path.relpath(fname)
            logger.info("Currently processing {}".format(fname))

            # if we want to save history, this is the output name it would take
            outname = os.path.join(
                os.path.dirname(fname), "." + os.path.basename(fname) + ".json"
            )

            # let's check if history already exists and load that file, if so
            if os.path.exists(outname):
                data = peakdet.load_history(outname)
            else:
                # load data with appropriate function, depending on source
                if source == "rtpeaks":
                    data = load_func(fname, fs=fs, channel=channel)
                else:
                    data = load_func(fname, fs=fs)

                # filter
                flims, method = MODALITIES[modality]
                data = peakdet.filter_physio(data, cutoffs=flims, method=method)

                # perform peak detection
                data = peakdet.peakfind_physio(data, thresh=thresh)

            # edit peaks, if desired (HIGHLY RECOMMENDED)
            # we'll do this even if we loaded from history
            # just to give another chance to check things over
            if not noedit:
                data = peakdet.edit_physio(data)

            # save back out to history, if desired
            if savehistory:
                peakdet.save_history(outname, data)

            # keep requested outputs
            hrv = peakdet.HRV(data)
            outputs = ["{:.5f}".format(getattr(hrv, attr, "")) for attr in measurements]

            # save as we go so that interruptions don't screw everything up
            dest.write(",".join([fname] + outputs) + "\n")


def main():
    logger.enable("")
    opts = get_parser().parse_args()
    workflow(**vars(opts))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Functions for loading and saving data and analyses
"""

import json
import os.path as op

import numpy as np
from loguru import logger

from peakdet import physio, utils

EXPECTED = ["data", "fs", "history", "metadata"]


def load_physio(data, *, fs=None, dtype=None, history=None, allow_pickle=False):
    """
    Returns `Physio` object with provided data

    Parameters
    ----------
    data : str or array_like or Physio_like
        Input physiological data. If array_like, should be one-dimensional
    fs : float, optional
        Sampling rate of `data`. Default: None
    dtype : data_type, optional
        Data type to convert `data` to, if conversion needed. Default: None
    history : list of tuples, optional
        Functions that have been performed on `data`. Default: None
    allow_pickle : bool, optional
        Whether to allow loading if `data` contains pickled objects. Default:
        False

    Returns
    -------
    data: :class:`peakdet.Physio`
        Loaded physiological data

    Raises
    ------
    TypeError
        If provided `data` is unable to be loaded
    """

    # first check if the file was made with `save_physio`; otherwise, try to
    # load it as a plain text file and instantiate a history
    if isinstance(data, str):
        try:
            inp = dict(np.load(data, allow_pickle=allow_pickle))
            for attr in EXPECTED:
                try:
                    inp[attr] = inp[attr].dtype.type(inp[attr])
                except KeyError:
                    raise ValueError(
                        "Provided npz file {} must have all of "
                        "the following attributes: {}".format(data, EXPECTED)
                    )
            # fix history, which needs to be list-of-tuple
            if inp["history"] is not None:
                inp["history"] = list(map(tuple, inp["history"]))
        except (IOError, OSError, ValueError):
            inp = dict(data=np.loadtxt(data), history=[utils._get_call(exclude=[])])
        logger.debug("Instantiating Physio object from a file")
        phys = physio.Physio(**inp)
    # if we got a numpy array, load that into a Physio object
    elif isinstance(data, np.ndarray):
        logger.debug("Instantiating Physio object from numpy array")
        if history is None:
            logger.warning(
                "Loading data from a numpy array without providing a"
                "history will render reproducibility functions "
                "useless! Continuing anyways."
            )
        phys = physio.Physio(np.asarray(data, dtype=dtype), fs=fs, history=history)
    # create a new Physio object out of a provided Physio object
    elif isinstance(data, physio.Physio):
        logger.debug(
            "Instantiating a new Physio object from the provided Physio object"
        )
        phys = utils.new_physio_like(data, data.data, fs=fs, dtype=dtype)
        phys._history += [utils._get_call()]
    else:
        raise TypeError("Cannot load data of type {}".format(type(data)))

    # reset sampling rate, as requested
    if fs is not None and fs != phys.fs:
        if not np.isnan(phys.fs):
            logger.warning(
                "Provided sampling rate does not match loaded rate. "
                "Resetting loaded sampling rate {} to provided {}".format(phys.fs, fs)
            )
        phys._fs = fs
    # coerce datatype, if needed
    if dtype is not None:
        phys._data = np.asarray(phys[:], dtype=dtype)

    return phys


def save_physio(fname, data):
    """
    Saves `data` to `fname`

    Parameters
    ----------
    fname : str
        Path to output file; .phys will be appended if necessary
    data : Physio_like
        Data to be saved to file

    Returns
    -------
    fname : str
        Full filepath to saved output
    """

    from peakdet.utils import check_physio

    data = check_physio(data)
    fname += ".phys" if not fname.endswith(".phys") else ""
    with open(fname, "wb") as dest:
        hist = data.history if data.history != [] else None
        np.savez_compressed(
            dest, data=data.data, fs=data.fs, history=hist, metadata=data._metadata
        )
    logger.info(f"Saved {data} in {fname}")

    return fname


def load_history(file, verbose=False):
    """
    Loads history from `file` and replays it, creating new Physio instance

    Parameters
    ----------
    file : str
        Path to input JSON file
    verbose : bool, optional
        Whether to print messages as history is being replayed. Default: False

    Returns
    -------
    file : str
        Full filepath to saved output
    """

    # import inside function for safety!
    # we'll likely be replaying some functions from within this module...
    import peakdet

    # grab history from provided JSON file
    with open(file, "r") as src:
        history = json.load(src)

    # replay history from beginning and return resultant Physio object
    logger.info(f"Replaying history from {file}")
    data = None
    for func, kwargs in history:
        if verbose:
            logger.info("Rerunning {}".format(func))
        # loading functions don't have `data` input because it should be the
        # first thing in `history` (when the data was originally loaded!).
        # for safety, check if `data` is None; someone could have potentially
        # called load_physio on a Physio object (which is a valid, albeit
        # confusing, thing to do)
        if "load" in func and data is None:
            if not op.exists(kwargs["data"]):
                if kwargs["data"].startswith("/"):
                    msg = (
                        "Perhaps you are trying to load a history file "
                        "that was generated with an absolute path?"
                    )
                else:
                    msg = (
                        "Perhaps you are trying to load a history file "
                        "that was generated from a different directory?"
                    )
                raise FileNotFoundError(
                    "{} does not exist. {}".format(kwargs["data"], msg)
                )
            data = getattr(peakdet, func)(**kwargs)
        else:
            data = getattr(peakdet, func)(data, **kwargs)

    return data


def save_history(file, data):
    """
    Saves history of physiological `data` to `file`

    Saved file can be replayed with `peakdet.load_history`

    Parameters
    ----------
    file : str
        Path to output file; .json will be appended if necessary
    data : Physio_like
        Data with history to be saved to file

    Returns
    -------
    file : str
        Full filepath to saved output
    """

    from peakdet.utils import check_physio

    data = check_physio(data)
    if len(data.history) == 0:
        logger.warning(
            "History of provided Physio object is empty. Saving "
            "anyway, but reloading this file will result in an "
            "error."
        )
    file += ".json" if not file.endswith(".json") else ""
    with open(file, "w") as dest:
        json.dump(data.history, dest, indent=4)
    logger.info(f"Saved {data} history in {file}")

    return file



"""
Utilities for testing
"""

from os.path import join as pjoin

import numpy as np
from pkg_resources import resource_filename

from peakdet import io, operations
from peakdet.utils import _get_call


def get_call_func(
    arg1,
    arg2,
    *,
    kwarg1=10,
    kwarg2=20,
    exclude=["exclude", "serializable"],
    serializable=True
):
    """Function for testing `peakdet.utils._get_call()`"""
    if arg1 > 10:
        kwarg1 = kwarg1 + arg1
    if arg2 > 20:
        kwarg2 = kwarg2 + arg2
    return _get_call(exclude=exclude, serializable=serializable)


def get_test_data_path(fname=None):
    """Function for getting `peakdet` test data path"""
    path = resource_filename("peakdet", "tests/data")
    return pjoin(path, fname) if fname is not None else path


def get_sample_data():
    """Function for generating tiny sine wave form for testing"""
    data = np.sin(np.linspace(0, 20, 40))
    peaks, troughs = np.array([3, 15, 28]), np.array([9, 21, 34])

    return data, peaks, troughs


def get_peak_data():
    """Function for getting some pregenerated physio data"""
    physio = io.load_physio(get_test_data_path("ECG.csv"), fs=1000)
    filt = operations.filter_physio(physio, [5.0, 15.0], "bandpass")
    peaks = operations.peakfind_physio(filt)

    return peaks



__all__ = [
    "delete_peaks",
    "edit_physio",
    "filter_physio",
    "interpolate_physio",
    "peakfind_physio",
    "plot_physio",
    "reject_peaks",
    "load_physio",
    "save_physio",
    "load_history",
    "save_history",
    "load_rtpeaks",
    "Physio",
    "HRV",
    "__version__",
]

from loguru import logger

from peakdet.analytics import HRV
from peakdet.external import load_rtpeaks
from peakdet.io import load_history, load_physio, save_history, save_physio
from peakdet.operations import (
    delete_peaks,
    edit_physio,
    filter_physio,
    interpolate_physio,
    peakfind_physio,
    plot_physio,
    reject_peaks,
)
from peakdet.physio import Physio

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

logger.disable("peakdet")



# -*- coding: utf-8 -*-
"""Functions and class for performing interactive editing of physiological data."""

import functools

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from loguru import logger
from matplotlib.widgets import SpanSelector
from packaging.version import Version

from peakdet import operations, utils


class _PhysioEditor:
    """
    Class for editing physiological data.

    Parameters
    ----------
    data : Physio_like
        Physiological data to be edited
    """

    def __init__(self, data):
        # save reference to data and generate "time" for interpretable X-axis
        self.data = utils.check_physio(data, copy=True)
        fs = 1 if data.fs is None else data.fs
        self.time = np.arange(0, len(data.data) / fs, 1 / fs)
        # Read if there is support data
        self.suppdata = data.suppdata

        # we need to create these variables in case someone doesn't "quit"
        # the plot appropriately (i.e., clicks X instead of pressing ctrl+q)
        self.deleted, self.rejected, self.included = set(), set(), set()

        # make main plot objects depending on supplementary data
        if self.suppdata is None:
            self.fig, self._ax = plt.subplots(
                nrows=1, ncols=1, tight_layout=True, sharex=True
            )
        else:
            self.fig, self._ax = plt.subplots(
                nrows=2,
                ncols=1,
                tight_layout=True,
                sharex=True,
                gridspec_kw={"height_ratios": [3, 2]},
            )

        self.fig.canvas.mpl_connect("scroll_event", self.on_wheel)
        self.fig.canvas.mpl_connect("key_press_event", self.on_key)

        # Set axis handler
        self.ax = self._ax if self.suppdata is None else self._ax[0]

        # three selectors for:
        #    1. rejection (central mouse),
        #    2. addition (right mouse), and
        #    3. deletion (left mouse)
        delete = functools.partial(self.on_edit, method="delete")
        reject = functools.partial(self.on_edit, method="reject")
        insert = functools.partial(self.on_edit, method="insert")

        # Check matplotlib version rectprops is deprecated with matplotlib 3.5.0 and then obsolete
        if Version(matplotlib.__version__) >= Version("3.5.0"):
            property_name = "props"
        else:
            property_name = "rectprops"

        self.span2 = SpanSelector(
            self.ax,
            delete,
            "horizontal",
            button=1,
            useblit=True,
            **{property_name: dict(facecolor="red", alpha=0.3)},
        )
        self.span1 = SpanSelector(
            self.ax,
            reject,
            "horizontal",
            button=2,
            useblit=True,
            **{property_name: dict(facecolor="blue", alpha=0.3)},
        )
        self.span3 = SpanSelector(
            self.ax,
            insert,
            "horizontal",
            button=3,
            useblit=True,
            **{property_name: dict(facecolor="green", alpha=0.3)},
        )

        self.plot_signals(False)

    def plot_signals(self, plot=True):
        """Clear axes and plots data / peaks / troughs."""
        # don't reset x-/y-axis zooms on replot
        if plot:
            xlim, ylim = self.ax.get_xlim(), self.ax.get_ylim()
        else:
            xlim, ylim = (-5, None), (None, None)

        # clear old data + redraw, retaining x-/y-axis zooms
        self.ax.clear()
        self.ax.plot(
            self.time,
            self.data,
            "b",
            self.time[self.data.peaks],
            self.data[self.data.peaks],
            ".r",
            self.time[self.data.troughs],
            self.data[self.data.troughs],
            ".g",
        )

        if self.suppdata is not None:
            self._ax[1].plot(self.time, self.suppdata, "k", linewidth=0.7)
            self._ax[1].set_ylim(-0.5, 0.5)

        self.ax.set(xlim=xlim, ylim=ylim, yticklabels="")
        self.fig.canvas.draw()

    def on_wheel(self, event):
        """Move axis on wheel scroll."""
        (xlo, xhi), move = self.ax.get_xlim(), event.step * -10
        self.ax.set_xlim(xlo + move, xhi + move)
        self.fig.canvas.draw()

    def quit(self):
        """Quit editor."""
        plt.close(self.fig)

    def on_key(self, event):
        """Undo last span select or quits peak editor."""
        # accept both control or Mac command key as selector
        if event.key in ["ctrl+z", "super+d"]:
            self.undo()
        elif event.key in ["ctrl+q", "super+d"]:
            self.quit()

    def on_edit(self, xmin, xmax, *, method):
        """
        Edit peaks by rejection, deletion, or insert.

        Removes specified peaks by either rejection / deletion, OR
        Include one peak by finding the max in the selection.

        method accepts 'insert', 'reject', 'delete'
        """
        logger.debug("Edited peaks with action: {}", method)
        if method not in ["insert", "reject", "delete"]:
            raise ValueError(f'Action "{method}" not supported.')

        tmin, tmax = np.searchsorted(self.time, (xmin, xmax))
        pmin, pmax = np.searchsorted(self.data.peaks, (tmin, tmax))

        if method == "insert":
            tmp = np.argmax(self.data.data[tmin:tmax]) if tmin != tmax else 0
            newpeak = tmin + tmp
            if newpeak == tmin:
                self.plot_signals()
                return
        else:
            bad = np.arange(pmin, pmax, dtype=int)
            if len(bad) == 0:
                self.plot_signals()
                return

        if method == "reject":
            rej, fcn = self.rejected, operations.reject_peaks
        elif method == "delete":
            rej, fcn = self.deleted, operations.delete_peaks

        # store edits in local history & call function
        if method == "insert":
            self.included.add(newpeak)
            self.data = operations.add_peaks(self.data, newpeak)
        else:
            rej.update(self.data.peaks[bad].tolist())
            self.data = fcn(self.data, self.data.peaks[bad])

        self.plot_signals()

    def undo(self):
        """Reset last span select peak removal."""
        # check if last history entry was a manual reject / delete
        relevant = ["reject_peaks", "delete_peaks", "add_peaks"]
        if self.data._history[-1][0] not in relevant:
            return

        # pop off last edit and delete
        func, peaks = self.data._history.pop()
        logger.debug(f"Undo previous action: {func}")

        if func == "reject_peaks":
            self.data._metadata["reject"] = np.setdiff1d(
                self.data._metadata["reject"], peaks["remove"]
            )
            self.rejected.difference_update(peaks["remove"])
        elif func == "delete_peaks":
            self.data._metadata["peaks"] = np.insert(
                self.data._metadata["peaks"],
                np.searchsorted(self.data._metadata["peaks"], peaks["remove"]),
                peaks["remove"],
            )
            self.deleted.difference_update(peaks["remove"])
        elif func == "add_peaks":
            self.data._metadata["peaks"] = np.delete(
                self.data._metadata["peaks"],
                np.searchsorted(self.data._metadata["peaks"], peaks["add"]),
            )
            self.included.remove(peaks["add"])
        self.data._metadata["troughs"] = utils.check_troughs(
            self.data, self.data.peaks, self.data.troughs
        )
        self.plot_signals()



# -*- coding: utf-8 -*-
"""
Various utilities for processing physiological data. These should not be called
directly but should support wrapper functions stored in `peakdet.operations`.
"""

import inspect
import sys
from functools import wraps

import numpy as np
from loguru import logger

from peakdet import physio


def make_operation(*, exclude=None):
    """
    Wrapper to make functions into Physio operations

    Wrapped functions should accept a :class:`peakdet.Physio` instance, `data`,
    as their first parameter, and should return a :class:`peakdet.Physio`
    instance

    Parameters
    ----------
    exclude : list, optional
        What function parameters to exclude from being stored in history.
        Default: 'data'
    """

    def get_call(func):
        @wraps(func)
        def wrapper(data, *args, **kwargs):
            # exclude 'data', by default
            ignore = ["data"] if exclude is None else exclude

            # grab parameters from `func` by binding signature
            name = func.__name__
            sig = inspect.signature(func)
            params = sig.bind(data, *args, **kwargs).arguments

            # actually run function on data
            data = func(data, *args, **kwargs)

            # it shouldn't be, but don't bother appending to history if it is
            if data is None:
                return data

            # get parameters and sort by key name, dropping ignored items and
            # attempting to coerce any numpy arrays or pandas dataframes (?!)
            # into serializable objects; this isn't foolproof but gets 80% of
            # the way there
            provided = {k: params[k] for k in sorted(params.keys()) if k not in ignore}
            for k, v in provided.items():
                if hasattr(v, "tolist"):
                    provided[k] = v.tolist()

            # append everything to data instance history
            data._history += [(name, provided)]

            return data

        return wrapper

    return get_call


def _get_call(*, exclude=None, serializable=True):
    """
    Returns calling function name and dict of provided arguments (name : value)

    Parameters
    ----------
    exclude : list, optional
        What arguments to exclude from provided argument : value dictionary.
        Default: ['data']
    serializable : bool, optional
        Whether to coerce argument values to JSON serializable form. Default:
        True

    Returns
    -------
    function: str
        Name of calling function
    provided : dict
        Dictionary of function arguments and provided values
    """

    exclude = ["data"] if exclude is None else exclude
    if not isinstance(exclude, list):
        exclude = [exclude]

    # get one function call up the stack (the bottom is _this_ function)
    calling = inspect.stack(0)[1]
    frame, function = calling.frame, calling.function

    # get all the args / kwargs from the calling function
    argspec = inspect.getfullargspec(frame.f_globals[function])
    args = sorted(argspec.args + argspec.kwonlyargs)

    # save arguments + argument values for everything not in `exclude`
    provided = {k: frame.f_locals[k] for k in args if k not in exclude}

    # if we want `provided` to be serializable, we can do a little cleaning up
    # this is NOT foolproof, but will coerce numpy arrays to lists which tends
    # to be the main issue with these sorts of things
    if serializable:
        for k, v in provided.items():
            if hasattr(v, "tolist"):
                provided[k] = v.tolist()

    return function, provided


def check_physio(data, ensure_fs=True, copy=False):
    """
    Checks that `data` is in correct format (i.e., `peakdet.Physio`)

    Parameters
    ----------
    data : Physio_like
    ensure_fs : bool, optional
        Raise ValueError if `data` does not have a valid sampling rate
        attribute.
    copy: bool, optional
        Whether to return a copy of the provided data. Default: False

    Returns
    -------
    data : peakdet.Physio
        Loaded physio object

    Raises
    ------
    ValueError
        If `ensure_fs` is set and `data` doesn't have valid sampling rate
    """

    from peakdet.io import load_physio

    if not isinstance(data, physio.Physio):
        data = load_physio(data)
    if ensure_fs and np.isnan(data.fs):
        raise ValueError("Provided data does not have valid sampling rate.")
    if copy is True:
        return new_physio_like(
            data, data.data, copy_history=True, copy_metadata=True, copy_suppdata=True
        )
    return data


def new_physio_like(
    ref_physio,
    data,
    *,
    fs=None,
    suppdata=None,
    dtype=None,
    copy_history=True,
    copy_metadata=True,
    copy_suppdata=True,
):
    """
    Makes `data` into physio object like `ref_data`

    Parameters
    ----------
    ref_physio : Physio_like
        Reference `Physio` object
    data : array_like
        Input physiological data
    fs : float, optional
        Sampling rate of `data`. If not supplied, assumed to be the same as
        in `ref_physio`
    suppdata : array_like, optional
        New supplementary data. If not supplied, assumed to be the same.
    dtype : data_type, optional
        Data type to convert `data` to, if conversion needed. Default: None
    copy_history : bool, optional
        Copy history from `ref_physio` to new physio object. Default: True
    copy_metadata : bool, optional
        Copy metadata from `ref_physio` to new physio object. Default: True
    copy_suppdata : bool, optional
        Copy suppdata from `ref_physio` to new physio object. Default: True

    Returns
    -------
    data : peakdet.Physio
        Loaded physio object with provided `data`
    """

    if fs is None:
        fs = ref_physio.fs
    if dtype is None:
        dtype = ref_physio.data.dtype
    history = list(ref_physio.history) if copy_history else []
    metadata = dict(**ref_physio._metadata) if copy_metadata else None

    if suppdata is None:
        suppdata = ref_physio._suppdata if copy_suppdata else None

    # make new class
    out = ref_physio.__class__(
        np.array(data, dtype=dtype),
        fs=fs,
        history=history,
        metadata=metadata,
        suppdata=suppdata,
    )
    return out


def check_troughs(data, peaks, troughs=None):
    """
    Confirms that `troughs` exists between every set of `peaks` in `data`

    Parameters
    ----------
    data : array-like
        Input data for which `troughs` and `peaks` were detected
    peaks : array-like
        Indices of suspected peak locations in `data`
    troughs : array-like or None, optional
        Indices of suspected troughs locations in `data`, if any.

    Returns
    -------
    troughs : np.ndarray
        Indices of trough locations in `data`, dependent on `peaks`
    """
    # If there's a trough after all peaks, keep it.
    if troughs is not None and troughs[-1] > peaks[-1]:
        all_troughs = np.zeros(peaks.size, dtype=int)
        all_troughs[-1] == troughs[-1]
    else:
        all_troughs = np.zeros(peaks.size - 1, dtype=int)

    for f in range(peaks.size - 1):
        dp = data[peaks[f] : peaks[f + 1]]
        idx = peaks[f] + np.argwhere(dp == dp.min())[0]
        all_troughs[f] = idx

    return all_troughs


def enable_logger(loglevel="INFO", diagnose=True, backtrace=True):
    """
    Toggles the use of the module's logger and configures it

    Parameters
    ----------
    loglevel : {'INFO', 'DEBUG', 'WARNING', 'ERROR'}
        Logger log level. Default: "INFO"
    """
    _valid_loglevels = ["INFO", "DEBUG", "WARNING", "ERROR"]

    if loglevel not in _valid_loglevels:
        raise ValueError(
            "Provided log level {} is not permitted; must be in {}.".format(
                loglevel, _valid_loglevels
            )
        )
    logger.enable("peakdet")
    try:
        logger.remove(0)
    except ValueError:
        logger.warning(
            "The logger has been already enabled. If you want to"
            "change the log level of an existing logger, please"
            "refer to the change_loglevel() function. (Note: You can"
            "find the log_handle either from the initial call of this"
            "function, or the console logs)"
        )
        return
    log_handle = logger.add(
        sys.stderr, level=loglevel, backtrace=backtrace, diagnose=diagnose
    )
    logger.debug(f"Enabling logger with handle_id: {log_handle}")
    return log_handle


def change_loglevel(log_handle, loglevel, diagnose=True, backtrace=True):
    """
    Change the loguru logger's log level. The logger needs to
    be already enabled by `enable_logger()`

    Parameters
    ----------
    log_handle : Enabled logger's handle, returned by `enable_logger()`
    loglevel : {'INFO', 'DEBUG', 'WARNING', 'ERROR'}
    """
    _valid_loglevels = ["INFO", "DEBUG", "WARNING", "ERROR"]

    if loglevel not in _valid_loglevels:
        raise ValueError(
            "Provided log level {} is not permitted; must be in {}.".format(
                loglevel, _valid_loglevels
            )
        )
    logger.remove(log_handle)
    new_log_handle = logger.add(
        sys.stderr, level=loglevel, backtrace=backtrace, diagnose=diagnose
    )
    logger.info(
        f'Changing the logger log level to "{loglevel}" (New logger handle_id: {new_log_handle})'
    )
    return new_log_handle


def disable_logger(log_handle=None):
    """
    Change the loguru logger's log level. The logger needs to
    be already enabled by `enable_logger()`

    Parameters
    ----------
    log_handle : Enabled logger's handle, returned by `enable_logger()`
        Default: None
        If left as None, this function will disable all logger instances
    """
    if log_handle is None:
        logger.info("Disabling all logger instances")
        logger.remove()
    else:
        logger.info(f"Disabling logger with handle_id: {log_handle}")
        logger.remove(log_handle)
    logger.disable("peakdet")



#!/usr/bin/env python
import sys

from setuptools import setup

import versioneer

SETUP_REQUIRES = ["setuptools >= 30.3.0"]
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

if __name__ == "__main__":
    setup(
        name="peakdet",
        setup_requires=SETUP_REQUIRES,
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
    )



import pytest
from _pytest.logging import LogCaptureFixture
from loguru import logger


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=20,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=False,
    )
    yield caplog
    logger.remove(handler_id)



# -*- coding: utf-8 -*-

import pytest

from peakdet import external
from peakdet.tests import utils as testutils

DATA = testutils.get_test_data_path("rtpeaks.csv")


def test_load_rtpeaks(caplog):
    for channel in [1, 2, 9]:
        hist = dict(fname=DATA, channel=channel, fs=1000.0)
        phys = external.load_rtpeaks(DATA, channel=channel, fs=1000.0)
        assert phys.history == [("load_rtpeaks", hist)]
        assert phys.fs == 1000.0
        with pytest.raises(ValueError):
            external.load_rtpeaks(
                testutils.get_test_data_path("ECG.csv"), channel=channel, fs=1000.0
            )
        assert caplog.text.count("WARNING") > 1



# -*- coding: utf-8 -*-

from peakdet import analytics
from peakdet.tests.utils import get_peak_data

ATTRS = [
    "rrtime",
    "rrint",
    "avgnn",
    "sdnn",
    "rmssd",
    "sdsd",
    "nn50",
    "pnn50",
    "nn20",
    "pnn20",
    "hf",
    "hf_log",
    "lf",
    "lf_log",
    "vlf",
    "vlf_log",
    "lftohf",
    "hf_peak",
    "lf_peak",
]


def test_HRV():
    peaks = get_peak_data()
    hrv = analytics.HRV(peaks)
    for attr in ATTRS:
        assert hasattr(hrv, attr)
